@@ -7,10 +7,18 @@ use App::sh2p::Parser;
 use App::sh2p::Here;
 
 sub App::sh2p::Parser::convert (\@\@);
-use constant (BREAK => '@');
+use constant (BREAK => 0x07);
 
 our $VERSION = '0.03';
 
+my %g_shell_options;
+
+########################################################
+#
+# Note for developers: 
+#      There are a lot of functions in here, 
+#      try to keep them in alphabetic order
+#
 ########################################################
 # For builtins/functionality that cannot be implemented
 sub not_implemented {
@@ -28,7 +36,7 @@ sub one4one {
     my $new_func = (App::sh2p::Parser::get_perl_builtin($func))[1];
     
     error_out ("$func replaced by Perl built-in $new_func\n".
-               " #               Check arguments and return value");
+               "#               Check arguments and return value");
     
     return general_arg_list($new_func, @args);
 }
@@ -52,7 +60,8 @@ sub general_arg_list {
     # Parse arguments
     for my $arg (@args) {           
         $ntok++;
-        if ($arg !~ /^\d$/) {
+
+        if ($arg !~ /^\d+$/) {
             # Escape embedded quotes
             $arg =~ s/\"/\\\"/g;
             #"help syntax highlighter
@@ -96,9 +105,34 @@ sub advise {
 
 ########################################################
 
+sub do_autoload {
+
+    my ($cmd, @functions) = @_;
+    my $ntok = 1;
+    
+    for my $func (@functions) {
+        my $first_char = substr($func,0,1);
+        last if $func eq BREAK || $func eq ';' || $first_char eq '#';
+
+        if ($first_char eq '$') {
+            # $cmd used - might be called from typedef
+            error_out "$cmd '$func' ignored";
+        }
+        else {
+            set_user_function ($func);
+        }
+        
+        $ntok++;
+    }
+    
+    return $ntok;   
+}
+
+########################################################
+
 sub do_break {
 
-   my ($undef, $level) = @_;
+   my (undef, $level) = @_;
    my $ntok = 1;
    
    iout 'last';
@@ -113,9 +147,26 @@ sub do_break {
 
 ########################################################
 
+sub do_colon {
+
+    my ($colon) = @_;
+    
+    if (!query_semi_colon()) {
+        # Probably in a conditional
+        out '(1)';
+    }
+    else {
+        iout '';
+    }
+    
+    return 1;
+}
+
+########################################################
+
 sub do_continue {
 
-   my ($undef, $level) = @_;
+   my (undef, $level) = @_;
    my $ntok = 1;
    
    iout 'next';
@@ -132,18 +183,13 @@ sub do_continue {
 
 sub do_cd {
 
-   my (undef, $directory) = @_;
+   my (undef, @rest) = @_;
    my $ntok = 1;
    
    iout 'chdir ';
+
+   $ntok += App::sh2p::Parser::join_parse_tokens ('.', @rest);
    
-   if (defined $directory && substr($directory,0,1) ne '#') {
-      my @tokens = ($directory);
-      my @types  = App::sh2p::Parser::identify (1, @tokens); 
-                        
-      App::sh2p::Parser::convert (@tokens, @types);
-      $ntok++;
-   }
    out ";\n";
    
    return $ntok;
@@ -202,6 +248,29 @@ sub do_chmod {
 
 ########################################################
 
+sub do_exec {
+
+   my (undef, @rest) = @_;
+   my $ntok = 1;
+   
+   if ($rest[0] =~ /^\d$/) {
+       error_out ("File descriptors are not supported");
+       $ntok = @_;
+       out "@_;\n";
+   }
+   else {
+       iout 'exec (';
+
+       $ntok += App::sh2p::Parser::join_parse_tokens (',', @rest);
+   
+       out ");\n";
+   }
+   
+   return $ntok;
+}
+
+########################################################
+
 sub do_exit {
 
    my ($cmd, $arg) = @_;
@@ -230,11 +299,15 @@ sub do_export {
    
    for my $env (@_) {
       if  ($env =~ /^(\w+)=(.*)$/) {
-         iout "\$ENV{$1} = $2;\n";
+         $env = $1;
+         iout "\$ENV{$env} = $2;\n";
       }
       else {
          iout "\$ENV{$env} = \$$env;\n";
+         iout "undef \$$env;\n";
+         Delete_variable ($env);
       }
+      Register_env_variable ($env);
    }
    
    return $ntok;
@@ -250,13 +323,31 @@ sub do_expr {
      
      # temporary fix
      error_out ('Suspious conversion from expr');
-     iout ("@rest");
+     
+     if (query_semi_colon()) {
+         iout ("@rest");
+     }
+     else {
+         out ("@rest");
+     }
      
      return $ntok; 
 }
 
 ########################################################
 
+sub do_functions {
+
+    my ($func) = @_;
+    
+    iout 'print map {"sub $_\n" if defined &{$_}} keys %main::;';
+    
+    return 1;
+
+}
+
+########################################################
+
 sub do_integer {
     
     # $cmd should be 'integer'
@@ -291,6 +382,98 @@ sub do_kill {
     return general_arg_list ($cmd, $signal, @rest);
 }
 
+########################################################
+# Also does for echo
+sub do_print {
+
+   my $ntok = 1;
+   my ($name, @args) = @_;
+   my $newline = 1;
+   my $handle = '';
+   
+   my $opt_u;
+   my %args;
+   local @ARGV;
+
+   my $redirection = '';
+   my $file = '';
+
+   # Move the comment to before the statement
+   if ( substr($args[-1],0,1) eq '#' ) {
+       my $comment = pop @args;
+       out "\n";
+       iout $comment;
+       $ntok++;
+   }
+   
+   for my $arg (@args) {
+       last if $arg eq BREAK || $arg eq ';';
+       
+       my $quotes = 0;
+       # This is so a > inside a string is not seen as redirection
+       if ($arg =~ /^([\"\']).*?\1/) {
+           $quotes = 1;
+       }
+       
+       # This should also strip out the redirection
+       if (!$quotes && $arg =~ s/(\>{1,2})//) {
+           $redirection = $1;     
+       }
+       
+       if ($arg && $redirection && (! $file)) {
+           $arg =~ s/(\S+)//;
+           $file = $1;
+       }
+       
+       push @ARGV, $arg if $arg;
+       $ntok++;
+   }
+   
+   if ($redirection) {
+       App::sh2p::Handlers::Handle_open_redirection ($redirection, $file);    
+       $handle = '$sh2p_handle ';
+   }
+   
+   getopts ('nEepu:', \%args);
+      
+   # Ignore -e and -E options (from echo)
+   if (exists $args{n}) {
+       $newline = 0;
+   }
+   
+   if ($name eq 'print') {
+       if (exists $args{p}) {
+           error_out ('Pipes/co-processes are not supported, use open');
+       }
+       
+       if (exists $args{u} && defined $args{u}) {
+           my @handles = ('', 'STDOUT ', 'STDERR ');
+           if ($args{u} > $#handles) {
+               error_out ('file descriptors not currently supported');
+               $handle = "$args{u} ";  # Just to show something 
+           }
+           else {
+               $handle = $handles[$args{u}];
+           }
+       }
+   }
+   
+   iout ("print $handle");
+   
+   App::sh2p::Parser::join_parse_tokens (',', @ARGV);
+      
+   if ($newline) {
+      out ",\"\\n\";\n"
+   }
+   else {
+      out ";\n";
+   }
+   
+   App::sh2p::Handlers::Handle_close_redirection() if $redirection;
+
+   return $ntok;
+}
+
 ########################################################
 
 sub do_read {
@@ -315,10 +498,10 @@ sub do_read {
    if (exists $args{p}) {               # Bash syntax
       $prompt = $args{p}
    }
-   
-   if ($ARGV[0] =~ /^(\w+)\?(.*)$/) {   # ksh syntax
+   elsif ($ARGV[0] =~ /^(\w*)\?(.*)$/) {   # ksh syntax
+       
+      $ARGV[0] = $1 || 'REPLY';    
       $prompt  = $2;
-      $ARGV[0] = $1;
    }   
 
    # Add $ prefix to variable names   
@@ -382,24 +565,70 @@ sub do_return {
 
 ########################################################
 
+sub do_shift {
+
+   my (undef, $level) = @_;
+   my $ntok = 1;
+   
+   if (defined $level && $level =~ /^\d+$/) {
+      $ntok++;
+   }
+   else {
+      $level = 1;
+   }
+
+   iout ('shift;' x $level);
+   
+   return $ntok;
+
+}
+
+########################################################
+
+sub do_shopt {
+
+   my (undef, $switch, @rest) = @_;
+   my $ntok = 2;
+   my @options;
+   
+   for my $option (@rest) {
+       last if $option eq BREAK || $option eq ';' || substr($option,0,1) eq '#';
+       push @options, $option;
+       $ntok++;
+   }  
+   
+   if ($switch eq '-s') {
+       @g_shell_options{@options} = undef;
+   }
+   elsif ($switch eq '+s') {
+       delete @g_shell_options{@options};
+   }
+   else {
+       error_out ("Unrecognised shopt argument: <$switch>");
+   }
+   
+   return $ntok;
+   
+}
+
+########################################################
+
 sub do_source {
 
-   my ($undef, $file) = @_;
+   my (undef, @tokens) = @_;
    my $ntok = 1;
    
    error_out ();
-   error_out "sourced file should also be converted\n";
-   out 'do ';
+   error_out "sourced file should also be converted";
+   iout 'do "';
+   
+   no_semi_colon(); 
+   
+   $ntok += App::sh2p::Parser::join_parse_tokens ('.', @tokens);
+   
+   reset_semi_colon();
+   out '";';
    
-   if (defined $file && substr($file,0,1) ne '#') {
-      my @tokens = ($file);
-      my @types  = App::sh2p::Parser::identify (1, @tokens); 
-                                                
-      App::sh2p::Parser::convert (@tokens, @types);
-      $ntok++;
-   }
-
-   out ";";
    return $ntok;
 }
 
@@ -440,6 +669,13 @@ sub do_typeset {
        if ( $args{Z} && defined $args{Z}) {
            $type =~ s/n/$args{Z}/;
        }
+       elsif ( $args{f} ) {
+           if ($args{u}) {
+               $ntok += do_autoload ('typeset -fu',@ARGV);
+               $ntok--;   # artificial 1st argument
+           }
+           return $ntok;
+       }
        else {
            error_out "Only one option supported for typedef or declare";
        }
@@ -459,95 +695,6 @@ sub do_typeset {
    return $ntok;
 }
 
-########################################################
-# Also does for echo
-sub do_print {
-
-   my $ntok = 1;
-   my ($name, @args) = @_;
-   my $newline = 1;
-   my $handle = '';
-   
-   my $opt_u;
-   my %args;
-   local @ARGV;
-
-   my $redirection = '';
-   my $file = '';
-
-   for my $arg (@args) {
-       last if $arg eq BREAK || $arg eq ';';
-       
-       my $quotes = 0;
-       # This is so a > inside a string is not seen as redirection
-       if ($arg =~ /^([\"\']).*?\1/) {
-           $quotes = 1;
-       }
-       
-       # This should also strip out the redirection
-       if (!$quotes && $arg =~ s/(\>{1,2})//) {
-           $redirection = $1;     
-       }
-       
-       if ($arg && $redirection && (! $file)) {
-           $arg =~ s/(\S+)//;
-           $file = $1;
-       }
-       
-       push @ARGV, $arg if $arg;
-       $ntok++;
-   }
-   
-   if ($redirection) {
-       App::sh2p::Handlers::Handle_open_redirection ($redirection, $file);    
-       $handle = '$sh2p_handle ';
-   }
-   
-   getopts ('nEepu:', \%args);
-      
-   # Ignore -e and -E options (from echo)
-   if (exists $args{n}) {
-       $newline = 0;
-   }
-   
-   if ($name eq 'print') {
-       if (exists $args{p}) {
-           error_out ('Pipes/co-processes are not supported, use open');
-       }
-       
-       if (exists $args{u} && defined $args{u}) {
-           my @handles = ('', 'STDOUT ', 'STDERR ');
-           if ($args{u} > $#handles) {
-               error_out ('file descriptors not currently supported');
-               $handle = "$args{u} ";  # Just to show something 
-           }
-           else {
-               $handle = $handles[$args{u}];
-           }
-       }
-   }
-   
-   iout ("print $handle");
-   
-   # C style for loop because I need to check the position
-   for (my $i; $i < @ARGV; $i++) {
-       my @args = ($ARGV[$i]);
-       my @types  = App::sh2p::Parser::identify (2, @args);                                
-       App::sh2p::Parser::convert (@args, @types);
-       out ',' if $i < $#ARGV;
-   }
-   
-   if ($newline) {
-      out ",\"\\n\";"
-   }
-   
-   out "\n";
-   
-   App::sh2p::Handlers::Handle_close_redirection() if $redirection;
-   
-   return $ntok;
-}
-
 ########################################################
 # Need getopt here, but it can't deal with +
 #  set [+-abCefhkmnpsuvxX] [+-o [option]] [+-A name] [--] [arg ...]
@@ -555,24 +702,41 @@ sub do_print {
 sub do_set {
    
    my $ntok = @_;
-   # First argument is 'read'
+   # First argument is 'set'
    shift @_;
+   my @values;
    
    for my $option (@_) {
       my $act = substr($option, 0, 1);
-      my $set = substr($option, 1);
+
+      if ($act eq '+' || $act eq '-') {
+          my $set = substr($option, 1);
       
-      if ( $set eq 'A') {
-         if ($act eq '-') {
-            initialise_array (@_);
-         }
-         else {
-            overwrite_array (@_);
-         }
+          if ( $set eq 'A') {
+              if ($act eq '-') {
+                 initialise_array (@_);
+              }
+              else {
+                 overwrite_array (@_);
+              }
          
-         last;
+              last;
+          }
+      }
+      else {
+          push @values, $option;
       }
    }  
+   
+   if (@values) {
+        iout "\@ARGV = ();\n";
+        iout "push \@ARGV,(";
+        
+        App::sh2p::Parser::join_parse_tokens (',', @values);
+                
+        out ");\n";
+   }
+   
    return $ntok;
 }
 
@@ -595,12 +759,30 @@ sub overwrite_array {
 
 sub do_unset {
    
-   my ($undef, $var) = @_;
+   my ($undef, $var, @rest) = @_;
    my $ntok = 1;
    
+   if (substr($var,0,1) eq '-') {
+       my $option = $var;
+    
+       $var = $rest[0];
+       $ntok++;
+      
+       # unset only supports two options (POSIX)
+       # -v has the same effect as not being there
+       
+       if ($option eq '-f') {
+           unset_user_function ($var);
+           $ntok++;
+           return $ntok;
+       }
+       
+   }
+   
    iout 'undef ';
 
    if (defined $var && substr($var,0,1) ne '#') {
+   
       my $type = '$';
       
       if (get_special_var($var)) {
