@@ -5,20 +5,21 @@ use strict;
 use Data::Dumper;   # for debug purposes
 
 sub App::sh2p::Parser::convert (\@\@);
-use constant (BREAK => '@');
+use constant (BREAK => 0x07);
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 require Exporter;
 our (@ISA, @EXPORT);
 @ISA = ('Exporter');
-@EXPORT = qw (Register_variable  Delete_variable    get_variable_type
+@EXPORT = qw (Register_variable  Register_env_variable
+              Delete_variable    get_variable_type
               print_types_tokens reset_globals
-              iout               out                error_out
-              get_special_var    set_special_var
+              iout               out                error_out    flush_out
+              get_special_var    set_special_var    can_var_interpolate
               mark_function      unmark_function    ina_function
               inc_block_level    dec_block_level    get_block_level
-              is_user_function   set_user_function
+              is_user_function   set_user_function  unset_user_function
               dec_indent         inc_indent 
               no_semi_colon      reset_semi_colon   query_semi_colon
               open_out_file      close_out_file);
@@ -28,20 +29,26 @@ our (@ISA, @EXPORT);
 my $g_indent_spacing = 4;
 
 my %g_special_vars = (
-      'IFS'   => '" \t\n"',
-      'ERRNO' => '\$!',
-      '?'     => '($? >> 8)',
-      '#'     => 'scalar(@ARGV)',
-      '@'     => '@ARGV',
-      '*'     => '@ARGV',    # Should do a join with IFS
-      '-'     => 'not supported',
-      '$'     => '$$',
-      '!'     => 'not supported'
+      'IFS'      => '" \t\n"',
+      'ERRNO'    => '$!',
+      'HOME'     => '$ENV{HOME}',
+      'PATH'     => '$ENV{PATH}',
+      'FUNCNAME' => 'caller(0))[3]',
+      '?'        => '($? >> 8)',
+      '#'        => 'scalar(@ARGV)',
+      '@'        => '@ARGV',
+      '*'        => '@ARGV',    # Should do a join with IFS
+      '-'        => 'not supported',
+      '$'        => '$$',
+      '!'        => 'not supported'
       );
 
 # This hash keeps a key for each variable declared
 # so we know if to put a 'my' prefix
-my %variables;
+my %g_variables;
+
+# This hash keeps track of environment variables
+my %g_env_variables;
 
 my %g_user_functions;
 my $g_new_line       = 1;
@@ -50,27 +57,52 @@ my $g_ina_function   = 0;
 my $g_block_level    = 0;
 my $g_indent         = 0;
 my $g_errors         = 0;
-my $g_line_number    = 1;
 
 my $g_outh;
 my $g_filename;
+my $g_out_buffer;
+my $g_err_buffer;
 
 #  For use by App::sh2p only
 ############################################################################
 
+sub can_var_interpolate {
+
+   my ($name) = @_;
+   my $retn;
+   
+   $retn = get_special_var ($name);
+   
+   if (defined $retn && $retn !~ /^[\$\@]/) {
+       return 0
+   }
+   else {
+       return 1
+   }
+}
+
+############################################################################
+
 sub get_special_var {
    my ($name) = @_;
    my $retn;
    
    return undef if ! defined $name;
-   
+
    # Remove dollar prefix
    $name =~ s/^\$//;
    
-   if ($name =~ /^(\d+)$/) {
+   if ($name eq '0') {
+       $retn = '$0';
+   }
+   elsif ($name =~ /^(\d+)$/) {
        my $offset = $1 - 1;
        $retn = "\$ARGV[$offset]";
    }
+   elsif ($name eq 'PWD') {
+       error_out ("Using \$PWD is unsafe: use Cwd::getcwd");
+       $retn = '$ENV{PWD}';
+   }
    else {
        $retn = $g_special_vars{$name};
    }
@@ -137,35 +169,50 @@ sub Register_variable {
     
     #print STDERR "Register_variable: <$name> <$type> <$level>\n";
     
-    if (exists $variables{$name}) {
+    if (exists $g_variables{$name}) {
     
-       if ($variables{$name}->[0] <= $level) {
+       if ($g_variables{$name}->[0] <= $level) {
            return 0
        }
        else {
-           # Possible change of type ?
+           # Create the variable with the block level and type	          
+           $g_variables{$name} = [$level, $type];
            return 1
        }
     }
+    elsif (exists $g_env_variables{$name}) {
+    
+       $g_env_variables{$name} = undef; 
+       return 0;
+    }
     else {
        # Create the variable with a block level and type
        
-       $variables{$name} = [$level, $type];
+       $g_variables{$name} = [$level, $type];
        return 1
     } 
 }
 
 ############################################################################
 
+sub Register_env_variable {
+    my ($name) = @_;
+    
+    # Does not matter if it already exists, or its type
+    $g_env_variables{$name} = undef; 
+}
+
+############################################################################
+
 sub get_variable_type {
 
     my ($name) = @_;
     my $level  = get_block_level();
         
-    if (exists $variables{$name}) {
+    if (exists $g_variables{$name}) {
   
-       if ($variables{$name}->[0] <= $level) {
-           return $variables{$name}->[1]
+       if ($g_variables{$name}->[0] <= $level) {
+           return $g_variables{$name}->[1]
        }
     }
     
@@ -173,14 +220,14 @@ sub get_variable_type {
 }
 
 ############################################################################
-# Called by unset
+# Called by unset and export
 sub Delete_variable {
     my ($name) = @_;
     my $level  = get_block_level();
         
-    if (exists $variables{$name}) {
-       if ($variables{$name} <= $level) {
-           delete $variables{$name}
+    if (exists $g_variables{$name}) {
+       if ($g_variables{$name} <= $level) {
+           delete $g_variables{$name}
        }
     }
    
@@ -195,8 +242,8 @@ sub inc_block_level {
 sub dec_block_level {
     
     # Remove registered variables of current block level
-    while (my($key, $value) = each (%variables)) {
-        delete $variables{$key} if $value == $g_block_level
+    while (my($key, $value) = each (%g_variables)) {
+        delete $g_variables{$key} if $value == $g_block_level
     }
     
     $g_block_level--;
@@ -214,16 +261,26 @@ sub get_block_level {
 
 sub is_user_function {
    my ($name) = @_;
+
    return (exists $g_user_functions{$name})
 }
 
 sub set_user_function {
    my ($name) = @_;
+
    $g_user_functions{$name} = undef;
    
    return 1;   # true
 }
 
+sub unset_user_function {
+   my ($name) = @_;
+   
+   delete $g_user_functions{$name} if exists $g_user_functions{$name};
+   
+   return 1;   # true
+}
+
 #################################################################################
 
 sub mark_new_line {
@@ -242,19 +299,27 @@ sub dec_indent { $g_indent-- if $g_indent > 0  }
 #################################################################################
 
 sub open_out_file {
-    $g_filename = shift;
+    my ($g_filename, $perms) = @_;
     
     if ($g_filename eq '-') {
         $g_outh = *STDOUT;
     }
     else {
         open ($g_outh, '>', $g_filename) || die "$g_filename: $!\n";
+        
+        # fchmod is not implemented on all platforms
+        chmod ($perms, $g_filename) if defined $perms;
         print STDERR "Processing $g_filename:\n";
     }
+    
+    $g_out_buffer = '';
+    $g_err_buffer = '';
 }
 
 sub close_out_file {
     
+    flush_out ();
+    
     close ($g_outh);
     print STDERR "\n";
     $g_filename = undef;
@@ -264,83 +329,101 @@ sub close_out_file {
 # Indented out
 sub iout {
 
-   print $g_outh ' ' x ($g_indent * $g_indent_spacing);
+   #print $g_outh ' ' x ($g_indent * $g_indent_spacing);
    
-   #if ($_[0] =~ /^my/) {
-   #    my @caller = caller();
-   #    print "iout my: @caller\n";
-   #}
+   my (@args) = @_;
+  
+   if (query_semi_colon()) {
+       unshift @args, (' ' x ($g_indent * $g_indent_spacing));
+   }
    
-   out (@_);
+   out (@args);
 }
 
 #################################################################################
 
 sub out {
-
-   #print STDERR "\nout <@_>\n";
-   #my @caller = caller();
-   #print STDERR "Called from @caller\n";
    
+   local $" = '';
    my $line = "@_";
    
-   while ($line =~ /\n/g) {
-       $g_line_number++
-   }
-   
-   print $g_outh $line;
-   
+   $g_out_buffer .= $line;
+      
    $g_new_line = 0;
+   
 }
 
-#################################################################################
+################################################################################
 
 sub error_out {
     my $msg = shift;
     
     if (defined $msg) {
         $msg = "**** INSPECT: $msg\n";
-        printf STDERR " %03d %s", $g_line_number, $msg;
     }
     else {
         $msg = "\n";
     }
-    
-    out "# $msg";
-    $g_line_number++;
+
+    $g_err_buffer .= "# $msg";
     
     $g_errors++;
 }
 
 #################################################################################
 
+sub flush_out {
+     
+   print $g_outh $g_err_buffer;
+   print $g_outh $g_out_buffer;
+      
+   # Leading space for readability with multiple files
+   $g_err_buffer =~ s/\#/ \#/g;
+   print STDERR $g_err_buffer; 
+   
+   $g_out_buffer = '';
+   $g_err_buffer = '';
+}
+
+#################################################################################
+
 sub reset_globals {
 
-    my %variables = ();
-
-    my %g_user_functions;
-    my $g_new_line       = 1;
-    my $g_use_semi_colon = 1;
-    my $g_ina_function   = 0;
-    my $g_block_level    = 0;
-    my $g_indent         = 0;
-    my $g_errors         = 0;
-    my $g_line_number    = 1;
+    %g_variables      = ();
+    %g_env_variables  = ();
+    %g_user_functions = ();
+    
+    $g_out_buffer     = '';
+    $g_err_buffer     = '';
+      
+    $g_new_line       = 1;
+    $g_use_semi_colon = 1;
+    $g_ina_function   = 0;
+    $g_block_level    = 0;
+    $g_indent         = 0;
+    $g_errors         = 0;
+    
 }
 
 #################################################################################
 # Debug purposes only
-sub print_types_tokens (\@\@) {
+sub print_types_tokens {
     
     my ($types, $tokens) = @_;
     
     for (my $i = 0; $i < @$types; $i++) {
     
-        print STDERR "Token: $tokens->[$i], type: $types->[$i][0]\n";
+        print STDERR "Type: ".$types->[$i][0].", ";
+        
+        print STDERR "Token: ".$tokens->[$i]."\n";
+    }
+    
+    if (@$types != @$tokens) {
+        print STDERR "Types array: ".@$types.", Token array: ".@$tokens."\n";
     }
 }
 
 #################################################################################
 
 # Module end
-1;
\ No newline at end of file
+1;
